subroutine rpn2(slice_direction,maxm,meqn,mwaves,mbc,mx,ql,qr,auxl,auxr,wave,s,amdq,apdq)

!--------------------------------------------------------------------------
! This solves Riemann problems for the thin film/surfactant model.
! The film (h=q(:,:,1)) and surfactant (Gamma=q(:,:,2)) act independently.
! On input, ql contains the state vector at the left edge of each cell,
! and qr contains the state vector at the right edge of each cell.
!
! The film advects with a Burgers-type equation,
!     h_t + (1/2 * h**2 * sigma_x)_x + (1/2 * h**2 * sigma_y)_y = 0.
!
! In b4step2, sigma_x and sigma_y have been stored on each cell, with
!
!     aux(i,j,1) = sigma_x(i,j), and
!     aux(i,j,2) = sigma_y(i,j).
!
!
! The surfactant evolves by simple advection,
!     Gamma_t + (-1/2 * beta * h^2 * h_x * Gamma)_x 
!             + (-1/2 * beta * h^2 * h_y * Gamma)_y = 0.
!
! The velocities have been stored in the aux arrays, with
!
!     aux(i,j,3) = -1/2 * beta * h^2 * h_x(i,j)
!     aux(i,j,4) = -1/2 * beta * h^2 * h_y(i,j).
!--------------------------------------------------------------------------


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: slice_direction, maxm, meqn, mwaves, mbc, mx
    double precision, intent(in) ::   ql(1-mbc:maxm+mbc, meqn)
    double precision, intent(in) ::   qr(1-mbc:maxm+mbc, meqn)
    double precision, intent(in) :: auxl(1-mbc:maxm+mbc, *)
    double precision, intent(in) :: auxr(1-mbc:maxm+mbc, *)

    !---- Outputs ----
    double precision, intent(out) :: wave(1-mbc:maxm+mbc, meqn, mwaves)
    double precision, intent(out) ::    s(1-mbc:maxm+mbc, mwaves)
    double precision, intent(out) :: apdq(1-mbc:maxm+mbc, meqn)
    double precision, intent(out) :: amdq(1-mbc:maxm+mbc, meqn)

    !---- Locals ----
    integer :: i
    double precision :: u_lower, u_upper

    !<=== Variables <===



    !===> Film height ===>

    do i=2-mbc,mx+mbc
        !---- Set velocity on either side of interface ----
        if (slice_direction==1) then
            u_lower = auxr(i-1,1)
            u_upper = auxl(i,1)
        else
            u_lower = auxr(i-1,2)
            u_upper = auxl(i,2)
        end if
        

        !---- This wave doesn't affect the surfactant ----
        wave(i,2,1) = 0.d0

        
        !---- Speed and fluctuations ----
        if ((u_lower>=0.d0) .and. (u_upper>=0.d0)) then

            !---- Both velocities are non-negative ----
            !-------------------------------------------------------------------
            ! All motion is to the right, so only the right-moving fluctuation
            ! apdq(i) is nonzero.  
            !
            ! The speed is determined by the Rankine-Hugoniot condition applied
            ! to the flux function
            !       f(q,x) = 1/2 * u(x) * q**2
            !-------------------------------------------------------------------

            wave(i,1,1) = u_upper*ql(i,1)**2/2.d0 - u_lower*qr(i-1,1)**2/2.d0
            s(i,1)      = .5d0 * (u_upper*ql(i,1) + sqrt(u_upper*u_lower)*qr(i-1,1))
            amdq(i,1)   = 0.d0
            apdq(i,1)   = wave(i,1,1)

        else if ((u_lower<=0.d0) .and. (u_upper<=0.d0)) then

            !---- Both velocities are negative ----

            wave(i,1,1) = u_upper*ql(i,1)**2/2.d0 - u_lower*qr(i-1,1)**2/2.d0
            s(i,1)      = .5d0 * (-sqrt(u_upper*u_lower)*ql(i,1) + u_lower*qr(i-1,1))
            amdq(i,1)   = wave(i,1,1)
            apdq(i,1)   = 0.d0

        else 
          
            !---- Velocities are in different directions ----
            !-------------------------------------------------------------------
            ! In this case, regardless of whether the velocities are moving
            ! towards or away from the interface, there is no flux across the
            ! interface itself.  However, to maintain conservation, amdq must
            ! take a value that compares the left state in cell i-1 to a right
            ! state of zero flux, and similarly for apdq regarding cell i.
            !-------------------------------------------------------------------

            wave(i,1,1) = 0.d0
            s(i,1)      = 0.d0    !(u_lower + u_upper)/2.d0 * (ql(i,1) + qr(i-1,1))/2.d0
            amdq(i,1)   = -u_lower * qr(i-1,1)**2/2.d0
            apdq(i,1)   = u_upper * ql(i,1)**2/2.d0

        end if
        
    end do
    
    !<=== Film height <===




    wave(:,:,2) = 0.d0
    s(:,2)      = 0.d0
    amdq(:,2)   = 0.d0
    apdq(:,2)   = 0.d0

!     !===> Surfactant concentration ===>
! 
!     do i = 2-mbc, mx+mbc
!         
!         !---- Select velocity for this cell interface ----          
!         if (slice_direction==1) then
!             u_lower = auxr(i-1,3)
!             u_upper = auxl(i,3)
!         else
!             u_lower = auxr(i-1,4)
!             u_upper = auxl(i,4)
!         end if
! 
! 
!         !---- This wave doesn't affect the film height ----
!         wave(i,1,2) = 0.d0
! 
!         
!         !---- Speed and fluctuations ----
!         if ((u_lower>=0.d0) .and. (u_upper>=0.d0)) then
! 
!             !---- Both velocities are non-negative ----
!             !------------------------------------------------------------------
!             ! All motion is to the right, so only the right-moving fluctuation
!             ! apdq(i) is nonzero.  The speed is determined by the lower cell,
!             ! as its mass is moving into the interface.
!             !------------------------------------------------------------------
! 
!             wave(i,2,2) = u_upper*ql(i,2) - u_lower*qr(i-1,2) 
!             s(i,2)      = u_upper
!             amdq(i,2)   = 0.d0
!             apdq(i,2)   = wave(i,2,2)
!             
!         else if ((u_lower<=0.d0) .and. (u_upper<=0.d0)) then
! 
!             !---- Both velocities are non-positive ----
!             !-----------------------------------------------------------------
!             ! All motion is to the left, so only the left-moving fluctuation
!             ! amdq(i) is nonzero.  The speed is determined by the upper cell,
!             ! as its mass is moving into the interface.
!             !-----------------------------------------------------------------
! 
!             wave(i,2,2) = u_upper*ql(i,2) - u_lower*qr(i-1,2) 
!             s(i,2)      = u_lower
!             amdq(i,2)   = wave(i,2,2)
!             apdq(i,2)   = 0.d0
! 
! 
!         else 
! 
!             !---- Velocities are in different directions ----
!             !-------------------------------------------------------------------
!             ! In this case, regardless of whether the velocities are moving
!             ! towards or away from the interface, there is no flux across the
!             ! interface itself.  However, to maintain conservation, amdq must
!             ! take a value that compares the left state in cell i-1 to a right
!             ! state of zero flux, and similarly for apdq regarding cell i.
!             !-------------------------------------------------------------------
! 
!             wave(i,2,2) = 0.d0
!             s(i,2)      = 0.d0    !(u_lower + u_upper) / 2.d0
!             amdq(i,2)   = -u_lower * qr(i-1,2)
!             apdq(i,2)   = u_upper * ql(i,2)
!             
!         end if        
!     
!     end do
!     
!     !<=== Surfactant concentration <===



    return
end subroutine rpn2