subroutine surfactant_parabolic_step( g, h, g_newton, dg, tolerance, success, &
                                      delta, mu, t, dt,                       &
                                      x_low, y_low, dx, dy, mx, my, mbc )

!----------------------------------------------------------
! Takes a time step with the parabolic portion of the 
! surfactant equation, using Witelski's Newton ADI method.
!----------------------------------------------------------

    !$ use omp_lib

    !===> Variables ===>

    implicit none

    !---- Inputs ----
    integer, intent(in) :: mx, my, mbc
    double precision, intent(in) :: h(1-mbc:mx+mbc,1-mbc:my+mbc)
    double precision, intent(in) :: tolerance, delta, mu, t, dt, x_low, y_low, dx, dy

    !---- In/out ----
    double precision, intent(inout) :: g(1-mbc:mx+mbc,1-mbc:my+mbc)

    !---- Output ----
    !-------------------------------------------------------------
    ! The output values of the arrays here are not actually used.
    ! The intent just indicates that their input values do not
    ! matter.
    !-------------------------------------------------------------
    double precision, intent(out), dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: g_newton, dg
    logical, intent(out) :: success
    
    !---- Globals ----
    integer :: newton_max_iter
    double precision :: newton_reduction_factor
    common /newton_block/ newton_max_iter, newton_reduction_factor
    !$ integer :: nthreads
    !$ common /omp_block/ nthreads

    !---- Locals ----
    double precision :: norm_dg, old_norm_dg
    integer :: iter, ix, iy, maxiter=10
    
    !---- External ----
    double precision, external :: film_true, surfactant_true
    
    !<==== Variables <===



    !---- Initialize ----
    !$ call omp_set_num_threads(nthreads)
    print *, "Entering surfactant_parabolic_step with dt = ", dt
    
    !$omp parallel do private(ix)
    do iy = 1,my
        do ix = 1,mx
            g_newton(ix,iy) = g(ix,iy)
        end do
    end do
    
    success  = .false.
    norm_dg     = tolerance + 1.d0
    old_norm_dg = norm_dg
    iter = 0
    
    
    !---- Fill ghost cells of g and h ----
    
    call fill_ghost_cells( h,        film_true,       t+dt, x_low, y_low, dx, dy, mx, my, mbc )
    call fill_ghost_cells( g,        surfactant_true, t   , x_low, y_low, dx, dy, mx, my, mbc )
    call fill_ghost_cells( g_newton, surfactant_true, t+dt, x_low, y_low, dx, dy, mx, my, mbc )
    
    
    

    !===> Newton iterations ===>

    do while (norm_dg > tolerance)
        
        !---- Iteration count ----
        
        iter = iter + 1
        
        
        !---- Calculate RHS for Newton's method; store in dg ----

        call surfactant_newton_rhs( dg, g_newton, g, h, &
                                    delta, mu, dx, dy, dt, mx, my, mbc )
        
        
        !---- Solve horizontal problems, overwriting dg ----

        !$omp parallel do
        do iy = 1,my
            call surfactant_integrate_slice( dg(1:mx,iy), g_newton(:,iy), h(:,iy), &
                                             delta, mu, dx, dt, mx, mbc )
        end do

        
        !---- Solve vertical problems, overwriting dg ----
        
        !$omp parallel do
        do ix = 1,mx
            call surfactant_integrate_slice( dg(ix,1:my), g_newton(ix,:), h(ix,:), &
                                             delta, mu, dy, dt, my, mbc )
        end do
        
        
        !---- Update g_newton, norm_dg, and old_norm_dg ----
        
        old_norm_dg = norm_dg
        norm_dg = 0.d0

        !$omp parallel do private(ix) reduction(max: norm_dg)
        do iy = 1,my
            do ix = 1,mx

                g_newton(ix,iy) = g_newton(ix,iy) + dg(ix,iy)

                norm_dg = max(norm_dg, abs(dg(ix,iy)))

            end do
        end do
        


        !---- Give up if norm_dg is too big ----
        
        if (norm_dg > 1e6) exit


        !---- If we've converged, then finish ----
        
        if (norm_dg < tolerance) then

            !$omp parallel do private(ix)
            do iy = 1,my
                do ix = 1,mx
                    g(ix,iy) = g_newton(ix,iy)
                end do
            end do

            success = .true.
            exit
        end if
        
        
        !---- If we've hit the max # of iterations, give up unless we're making sufficient progress ----

        if ( (iter >= newton_max_iter) .and. (norm_dg > newton_reduction_factor*old_norm_dg) ) exit

    end do
    !<=== Newton iterations <===




end subroutine surfactant_parabolic_step