subroutine periodic_banded_solve ( a_banded, b, n1, n2, n_lower, n_upper )

!-------------------------------------------------------------------------------
! This solves a system involving a "periodic banded matrix", as arises
! in discretization of a periodic PDE.  The matrix A is essentially banded, 
! except each lower diagonal wraps into the upper-right corner of the matrix,
! and each upper diagonal wraps around to the lower left.  As a simple example,
!
!      M1 U2 V3  0  0 L6
!      L1 M2 U3 V4  0  0
!       0 L2 M3 U4 V5  0
!       0  0 L3 M4 U5 V6
!      V1  0  0 L4 M5 U6
!      U1 V2  0  0 L5 M6.
!
! From the perspective of diagonals wrapping around the matrix, we extend
! LAPACK's usual sparse storage scheme, simply filling in the elements
! that are typically unused with the wrapped portions of the diagonals.  The
! matrix pictured above would be stored as
!
!      V1 V2 V3 V4 V5 V6
!      U1 U2 U3 U4 U5 U6
!      M1 M2 M3 M4 M5 M6
!      L1 L2 L3 L4 L5 L6
!
! with one (more generally, the number of lower diagonals) row of zeros as a
! workspace.  This has the pleasant effect of preserving column indices in
! the full-to-sparse mapping.
!
! The solution procedure is based on a block decomposition
!
!    A_11 | A_12
!   ------+------
!    A_21 | A_22,
!
! in which A_11 is purely banded.  This has been referred to elsewhere as
! a "border banded" matrix, though the terminology appears nonstandard.
! The solution procedure is straightforward, and one may step through the
! routine border_banded_solve for more details.  The more difficult issue
! is assembling the "border blocks" A_12, A_21, and A_22 from the banded
! storage scheme, and this is performed in subroutine assemble_border_blocks.
!-------------------------------------------------------------------------------


    implicit none
    
    integer, intent(in) :: n1, n2, n_lower, n_upper
    double precision, intent(inout) :: a_banded(2*n_lower+n_upper+1,n1+n2)
    double precision, intent(inout) :: b(n1+n2)
    
    double precision :: a_12(n1,n2), a_21(n2,n1), a_22(n2,n2)
    integer :: n_bands




    call assemble_border_blocks( a_banded, a_21, a_12, a_22, n1, n2, n_lower, n_upper)


    n_bands = n_lower + n_upper + 1
    
    call border_banded_solve( a_banded(1:n_lower+n_bands,1:n1), a_21, a_12, a_22, b, &
                              n1, n2, n_lower, n_upper )


end subroutine periodic_banded_solve





subroutine border_banded_solve ( a_11_banded, a_21, a_12, a_22, b, n1, n2, n_lower, n_upper )

    implicit none
    
    !==== Variables ====
    !---- Input ----
    integer, intent(in) :: n1, n2, n_lower, n_upper
    
    !---- Input/output ---
    double precision, intent(inout) :: a_11_banded(2*n_lower+n_upper+1, n1)
    double precision, intent(inout) :: a_21(n2,n1), a_12(n1,n2), a_22(n2,n2)
    double precision, intent(inout) :: b(n1+n2)

    !---- Locals ----
    integer :: pivots_1(n1), pivots_2(n2), info
    integer :: n
    
    
    !---- Full size of the system ----
    n = n1 + n2
    
    
    !---- Factorize a_11 ----
    call dgbtrf( n1, n1, n_lower, n_upper, a_11_banded, size(a_11_banded,1), pivots_1, info)


    !---- b(1:n1) = inv(a_11)*b(1:n1) ----
    call dgbtrs( 'No transpose', n1, n_lower, n_upper, 1, &
                 a_11_banded, size(a_11_banded,1), pivots_1, b(1:n1), n1, info )
    

    !---- a_12 maps to inv(a_11)*a_12 ----
    call dgbtrs( 'No transpose', n1, n_lower, n_upper, size(a_12,2), &
                 a_11_banded, size(a_11_banded,1), pivots_1, a_12, size(a_12,1), info )


    !---- a_22 = a_22 - a_21*a_12 ----
    a_22 = a_22 - matmul(a_21,a_12)
    
    
    !---- b_2 = b_2 - a_21*b_1 ----
    b(n1+1:n) = b(n1+1:n) - matmul(a_21, b(1:n1))
    
    
    !---- Solve: x_2 = inv(a_22)*b_2, overwriting b_2 ----
    call dgesv( n2, 1, a_22, n2, pivots_2, b(n1+1:n), n2, info )
    

    !---- Solve: x_1 = b_1 - a_12*x_2, overwriting b_1 ----
    b(1:n1) = b(1:n1) - matmul(a_12, b(n1+1:n))


end subroutine border_banded_solve




subroutine assemble_border_blocks ( a_banded, a_21, a_12, a_22, n1, n2, n_lower, n_upper )

    implicit none
    
    !==== Variables ====
    !---- Inputs ----
    integer, intent(in) :: n1, n2, n_lower, n_upper
    double precision, intent(in) :: a_banded(2*n_lower+n_upper+1, n1+n2)

    !---- Input/output ----
    double precision, intent(inout) :: a_12(n1,n2), a_21(n2,n1), a_22(n2,n2)

    !---- Locals ----
    integer :: n_to_copy, j, j_block, i_block, i_banded_start, i_banded_end
    integer :: n_zeros_upper, n_zeros_lower, n_nonzeros, n_bands
    
    
    
    n_bands = n_lower + n_upper + 1
    
    
    
    !===> Fill A_21 (bottom block) ===>

    a_21(:,:) = 0.d0

    !---- Copy upper diagonal elements in first n_upper columns ----
    n_to_copy = n_upper

    do j = 1, n_upper

        a_21( n2-n_to_copy+1:n2, j ) = a_banded( n_lower+1:n_lower+n_to_copy, j )

        n_to_copy = n_to_copy - 1

    end do


    !---- Copy lower diagonal elements in columns n1-n_lower+1 through n1
    n_to_copy = 0

    do j = n1-n_lower+1, n1

        n_to_copy = n_to_copy + 1

        a_21( 1:n_to_copy, j ) = a_banded( n_lower+n_bands-n_to_copy+1:n_lower+n_bands, j )

    end do

    !<=== Fill A_21 (bottom block) <===




    !===> Fill A_12 ===>

    a_12(:,:) = 0.d0

    do j_block = 1, n2

        !---- Column indices of each storage space ----
        j = n1 + j_block


        !---- Copy elements from lower diagonals ----
        n_to_copy = n_lower - n2 + j_block

        if (n_to_copy > 0) then
            a_12( 1:n_to_copy, j_block ) = a_banded(n_lower+n_bands-n_to_copy+1:n_lower+n_bands, j)
        end if

        !---- Copy elements from upper diagonals ----
        n_to_copy = n_upper + 1 - j_block

        if (n_to_copy > 0) then
            a_12( n1-n_to_copy+1:n1, j_block ) = a_banded(n_lower+1:n_lower+n_to_copy, j)
        end if

    end do

    !<=== Fill A_12 <===



    !===> Fill A_22 ===>

    i_banded_start = n_lower + n_upper + 1
    i_banded_end   = n_lower + n_upper + n2

    do j_block = 1, n2

        i_block = 1

        j = n1 + j_block

        n_zeros_upper = max(n_lower+1 - i_banded_start, 0)
        n_zeros_lower = max(i_banded_end - (n_lower+n_bands), 0)
        n_nonzeros    = n2 - n_zeros_upper - n_zeros_lower


        !---- Fill with zeros for missing upper diagonals, if necessary ----
        if (n_zeros_upper > 0) then
            a_22( i_block:i_block+n_zeros_upper-1, j_block ) = 0.d0
            i_block = i_block + n_zeros_upper
        end if


        !---- Copy nonzero elements ---
        a_22( i_block:i_block+n_nonzeros-1, j_block ) &
            = a_banded( i_banded_start+n_zeros_upper:i_banded_end-n_zeros_lower, j )

        i_block = i_block + n_nonzeros


        !---- Fill with zeros for missing lower diagonals, if necessary ----
        if (n_zeros_lower > 0) then
            a_22( i_block:i_block+n_zeros_lower-1, j_block ) = 0.d0
        end if


        !---- Update start and end indices for a_banded ----
        i_banded_start = i_banded_start - 1
        i_banded_end   = i_banded_end   - 1

    end do

    !<=== Fill A_22 <===
    
    
    
    
    
end subroutine assemble_border_blocks