subroutine film_integrate_slice_periodic(rhs, h, h_laplacian, beta, kappa, dx, dt, mx)

!---------------------------------------------------------------
! Solves a 1-dimensional problem for the approximately-factored
! thin film equation.
!
! The primary work of this routine is unpacking the sparsely
! stored Jacobian of the flux function into 'backward_euler',
! which is passed to LAPACK routine 'dgbsv' to solve the system.
!
! The first two rows of backward_euler are working space for
! dgbsv, and we do not fill them.  Rows 3 through 7 store the
! five diagonals of the system's "Jacobian" (which is actually
! an approximate factor of the full Jacobian for the 2D system).
! The column index of each element remains fixed in transition
! to the row-based form, as illustrated below for a 7x7 matrix:
!
!   Full matrix:
!
!         c1 b2 a3
!         d1 c2 b3 a4
!         e1 d2 c3 b4 a5
!            e2 d3 c4 b5 a6 
!               e3 d4 c5 b6 a7
!                  e4 d5 c6 b7
!                     e5 d6 c7
!
!
!   Rows 3 through 7 of backward_euler:
!
!          0  0 a3 a4 a5 a6 a7
!          0 b2 b3 b4 b5 b6 b7
!         c1 c2 c3 c5 c5 c6 c7
!         d1 d2 d3 d4 d5 d6  0
!         e1 e2 e3 e4 e5  0  0
!
!---------------------------------------------------------------


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: h(mx), h_laplacian(mx)
    double precision, intent(in) :: beta, kappa, dx, dt
    !---- In/out ----
    double precision, intent(inout) :: rhs(mx)
    !---- Locals ----
    integer :: ipiv(mx), info, i, j, n_lower, n_upper
    parameter (n_lower=2, n_upper=2)
    double precision :: backward_euler(2*n_lower+1+n_upper, mx), sparse_flux(n_lower+n_upper,mx)

    !<=== Variables <===



    !---- Set sparse_flux ----
    call film_sparse_flux_periodic( sparse_flux, h, h_laplacian, beta, kappa, dx, dt, mx )



    !===> Define backward_euler ===>

!     do j = 1, mx
! 
!         backward_euler(n_lower+1,j) = dt/dx * sparse_flux(1,j)
!         
!         do i_flux = 2, n_lower+n_upper-1
!             backward_euler(n_lower+1+i_flux,j) = dt/dx * (sparse_flux(i_flux,j) - sparse_flux(i_flux-1,j))
!         end do
!         
!         backward_euler(2*n_lower+1,j) = backward_euler(2*n_lower+1,j) + 1.d0
!         
!         backward_euler(2*n_lower+n_upper+1,j) = dt/dx * ( -sparse_flux(n_lower+n_upper,j) )
! 
!     end do
    

    !---- Rows 3:7 hold diagonals of the Jacobian ----
    !---- Column 1 ----
    backward_euler(3,1) =        dt/dx * (sparse_flux(1,1)                   )
    backward_euler(4,1) =        dt/dx * (sparse_flux(2,1) - sparse_flux(1,1))
    backward_euler(5,1) = 1.d0 + dt/dx * (sparse_flux(3,1) - sparse_flux(2,1))
    backward_euler(6,1) =        dt/dx * (sparse_flux(4,1) - sparse_flux(3,1))
    backward_euler(7,1) =        dt/dx * (                 - sparse_flux(4,1))


    !---- Column 2: Influence of cell 2 on cells 1 through 4 ----
    backward_euler(3,2) =        dt/dx * (sparse_flux(1,2)                   )
    backward_euler(4,2) =        dt/dx * (sparse_flux(2,2) - sparse_flux(1,2))
    backward_euler(5,2) = 1.d0 + dt/dx * (sparse_flux(3,2) - sparse_flux(2,2))
    backward_euler(6,2) =        dt/dx * (sparse_flux(4,2) - sparse_flux(3,2))
    backward_euler(7,2) =        dt/dx * (                 - sparse_flux(4,2))


    !---- Interior ----
    do j = 3,mx-2

        backward_euler(3,j) =        dt/dx * (sparse_flux(1,j)                   )
        backward_euler(4,j) =        dt/dx * (sparse_flux(2,j) - sparse_flux(1,j))
        backward_euler(5,j) = 1.d0 + dt/dx * (sparse_flux(3,j) - sparse_flux(2,j))
        backward_euler(6,j) =        dt/dx * (sparse_flux(4,j) - sparse_flux(3,j))
        backward_euler(7,j) =        dt/dx * (                 - sparse_flux(4,j))

    end do


    !---- Column mx-1 ----
    backward_euler(3,mx-1) =        dt/dx * (sparse_flux(1,mx-1)                      )
    backward_euler(4,mx-1) =        dt/dx * (sparse_flux(2,mx-1) - sparse_flux(1,mx-1))
    backward_euler(5,mx-1) = 1.d0 + dt/dx * (sparse_flux(3,mx-1) - sparse_flux(2,mx-1))
    backward_euler(6,mx-1) =        dt/dx * (sparse_flux(4,mx-1) - sparse_flux(3,mx-1))
    backward_euler(7,mx-1) =        dt/dx * (                    - sparse_flux(4,mx-1))


    !---- Column mx ----
    backward_euler(3,mx) =        dt/dx * (sparse_flux(1,mx)                    )
    backward_euler(4,mx) =        dt/dx * (sparse_flux(2,mx) - sparse_flux(1,mx))
    backward_euler(5,mx) = 1.d0 + dt/dx * (sparse_flux(3,mx) - sparse_flux(2,mx))
    backward_euler(6,mx) =        dt/dx * (sparse_flux(4,mx) - sparse_flux(3,mx))
    backward_euler(7,mx) =        dt/dx * (                  - sparse_flux(4,mx))

    !<=== Define backward_euler <===



    !---- Solve; rhs is overwritten with the result ----
!     call dgbsv(mx, n_lower, n_upper, 1, backward_euler, 2*n_lower+n_upper+1, ipiv, rhs, mx, info)
    call periodic_banded_solve( backward_euler, rhs, mx-2, 2, 2, 2 )


end subroutine film_integrate_slice_periodic