subroutine film_sparse_flux_periodic(sparse_flux, h, h_laplacian, beta, kappa, dx, dt, mx)

!-------------------------------------------------------------
! This routine calculates the components in the 1-dimensional
! flux operator in an approximate factorization of the
! Jacobian for Newton's method.
!
! Factor of 1/2 comes from multiplicative factors.  Since this
! is  an interface-centered operator, that takes an average
! of cell-centered values from either side.
!-------------------------------------------------------------


    !===> Variables ===>

    implicit none
    
    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: h(mx), h_laplacian(mx)
    double precision, intent(in) :: beta, kappa, dx, dt

    !---- Outputs ----
    double precision, intent(out) :: sparse_flux(4,mx)
    

    !---- Locals ----
    double precision :: h_face, h_x, h_laplacian_x, f, fd1
    integer :: j

    !<==== Variables <====



    !===> Lower boundary ===>
    
    !---- Interface 1 ----
    
    h_face        = (h(mx) + h(1)) / 2.d0
    h_x           = (h(1) - h(mx)) / dx
    h_laplacian_x = (h_laplacian(1) - h_laplacian(mx)) / dx
    f             = h_face**3 / 3.d0
    fd1           = h_face**2
    
    sparse_flux(4,mx-1) = f *(-kappa/dx**3)
    sparse_flux(3,mx)   = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*( beta/dx + 3.d0*kappa/dx**3)
    sparse_flux(2,1)    = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
    sparse_flux(1,2)    = f*(kappa/dx**3)


    !---- Interface 2 ----
        
    h_face        = (h(1) + h(2)) / 2.d0
    h_x           = (h(2) - h(1)) / dx
    h_laplacian_x = (h_laplacian(2) - h_laplacian(1)) / dx
    f             = h_face**3 / 3.d0
    fd1           = h_face**2

    sparse_flux(4,mx) = f*(-kappa/dx**3)
    sparse_flux(3,1)  = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*( beta/dx + 3.d0*kappa/dx**3)
    sparse_flux(2,2)  = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
    sparse_flux(1,3)  = f*(kappa/dx**3)

    !<=== Lower boundary <===



    !===> Interior interfaces ===>

    do j = 3,mx-1
  
        h_face        = (h(j-1) + h(j)) / 2.d0
        h_x           = (h(j) - h(j-1)) / dx
        h_laplacian_x = (h_laplacian(j) - h_laplacian(j-1)) / dx
        f             = h_face**3 / 3.d0
        fd1           = h_face**2
            
        sparse_flux(4,j-2) = f*(-kappa/dx**3)
        sparse_flux(3,j-1) = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(beta/dx + 3.d0*kappa/dx**3)
        sparse_flux(2,j)   = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
        sparse_flux(1,j+1) = f*(kappa/dx**3)

    end do

    !<=== Interior interfaces <===    



    !===> Upper boundary ===>

    !---- Interface mx ----
    !----------------------------------------------------------------
    ! Like element (3,1), the component in (2,mx) is affected by the 
    ! even extension across the boundary, which sets h(mx+1)=h(mx).
    !----------------------------------------------------------------
        
    h_face        = (h(mx-1) + h(mx)) / 2.d0
    h_x           = (h(mx) - h(mx-1)) / dx
    h_laplacian_x = (h_laplacian(mx) - h_laplacian(mx-1)) / dx
    f             = h_face**3 / 3.d0
    fd1           = h_face**2
    
    sparse_flux(4,mx-2) = f * (-kappa/dx**3)
    sparse_flux(3,mx-1) = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(beta/dx + 3.d0*kappa/dx**3)
    sparse_flux(2,mx)   = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
    sparse_flux(1,1)    = f*(kappa/dx**3)

    !<=== Upper boundary <===



end subroutine film_sparse_flux_periodic