subroutine surfactant_newton_rhs(rhs, g, g_old, h_old, delta, mu, dx, dy, dt, mx, my, mbc)

!===> Description ===>
!
! Evaluates the RHS of Newton's method for the parabolic part of
! the surfactant equation.
!
!<=== Description <===


    !$ use omp_lib
    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx, my, mbc
    double precision, intent(in), dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: g, g_old, h_old
    double precision, intent(in) :: delta, mu, dx, dy, dt

    !---- Outputs ----
    double precision, intent(out), dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: rhs

    !---- Locals ----
    integer :: ix, iy
    double precision :: h_face, g_face
    double precision :: tension_d1(mx+1), flux(mx+1)
    
    !$ integer :: nthreads
    !$ common /omp_block/ nthreads

    !<=== Variables <===


    !$ call omp_set_num_threads(nthreads)
    
    

    !---- First part of the RHS is easy ----
    rhs(1:mx,1:my) = g_old(1:mx,1:my) - g(1:mx,1:my)


    !===> Fluxing at fixed-x interfaces ===>
    !$omp parallel do private(ix, h_face, g_face, flux, tension_d1)
    do iy = 1,my
      
        !---- Calculate surface tension derivative at interfaces ----
        !----------------------------------------------------------
        ! Function/subroutine calls involve significant overhead.
        ! Thus, surface_tension_d1 is called only once for the
        ! whole one-dimensional section rather than once for each
        ! cell, even though the latter would be more natural in
        ! terms of organization.
        !----------------------------------------------------------
        
        do ix = 1,mx+1
            tension_d1(ix) = (g(ix-1,iy) + g(ix,iy)) / 2.d0
        end do
        call surface_tension_d1(tension_d1, mu, mx+1)


        !---- Calculate fluxes ----

        do ix = 1, mx+1
            h_face = (h_old(ix-1,iy) + h_old(ix,iy)) / 2.d0
            g_face = (g(ix-1,iy) + g(ix,iy)) / 2.d0
            
            flux(ix) = (h_face*g_face*tension_d1(ix) - delta) * (g(ix,iy) - g(ix-1,iy)) / dx
        end do
        
        
        !---- Update RHS ----
        
        do ix = 1, mx
            rhs(ix,iy) = rhs(ix,iy) - dt/dx * (flux(ix+1) - flux(ix))
        end do
        
    end do
    
    !<=== Fluxing at fixed-x interfaces <===
    
    
    
    !===> Fluxing at fixed-y interfaces ===>
    
    !$omp parallel do private(ix, h_face, g_face, flux, tension_d1)
    do iy = 1, my+1

        !---- Calculate surface tension derivative at interfaces ----
        
        do ix = 1,mx
            tension_d1(ix) = (g(ix,iy-1) + g(ix,iy)) / 2.d0
        end do
        call surface_tension_d1(tension_d1, mu, mx)
        
        
        !---- Compute fluxes ----
        
        do ix = 1, mx
            h_face = (h_old(ix,iy-1) + h_old(ix,iy)) / 2.d0
            g_face = (g(ix,iy-1) + g(ix,iy)) / 2.d0

            flux(ix) = (h_face*g_face*tension_d1(ix) - delta) * (g(ix,iy) - g(ix,iy-1)) / dy
        end do
        

        !---- Update the RHS based on fluxes on this section ----        

        if (iy > 1) then
            do ix = 1,mx          
                rhs(ix,iy-1) = rhs(ix,iy-1) - dt/dy * flux(ix)
            end do
        end if

        if (iy <= my) then
            do ix = 1, mx
                rhs(ix,iy) = rhs(ix,iy) + dt/dy * flux(ix)
            end do
        end if
        
    end do
    !<=== Fluxing at fixed-y interfaces <===


end subroutine surfactant_newton_rhs
