subroutine b4step2(maxmx, maxmy, mbc, mx, my, meqn, q,  &
                   x_low, y_low, dx, dy, t, dt, maux, aux)

!-----------------------------------------------------------------
! Called from claw2 before each call to step2.  Used to set the 
! aux arrays with velocity field data.
!
!    aux(:,:,1) = film velocity field, x-component
!    aux(:,:,2) = film velocity field, y-component
!    aux(:,:,3) = surfactant velocity field, x-component
!    aux(:,:,4) = surfactant velocity field, y-component
!-----------------------------------------------------------------

    !$ use omp_lib
    implicit none

    !---> Variables --->
    ! Input
    integer, intent(in) :: maxmx, maxmy, mbc, mx, my, meqn, maux
    double precision, intent(in) :: x_low, y_low, dx, dy, t, dt
    double precision, dimension(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc, meqn),  &
        intent(in) :: q

    ! Output
    double precision, dimension(1-mbc:maxmx+mbc,1-mbc:maxmy+mbc, maux),  &
        intent(out) :: aux

    ! Global
    double precision :: beta, kappa, mu, delta
    common /param_block/ beta, kappa, mu, delta

    ! Local
    integer :: ix, iy
    double precision, dimension(mx) :: tension_d1
    double precision, dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: film_laplacian
    double precision :: film_x, film_y, film_laplacian_x, film_laplacian_y,  &
        surfactant_x, surfactant_y
    !$ integer :: nthreads
    !$ common /omp_block/ nthreads
    !<--- Variables <---
    
    !$ call omp_set_num_threads(nthreads)
    
    !---> Film velocity field, x-component --->
    ! Fill interior cells
    !$omp parallel do private(tension_d1, ix, surfactant_x)
    do iy=1,my
        
        ! Set surface tension on horizontal slice
        tension_d1(1:mx) = q(1:mx,iy,2)
        call surface_tension_d1(tension_d1,mu,mx)
      
        do ix=1,mx
            surfactant_x = (q(ix+1,iy,2) - q(ix-1,iy,2)) / (2.d0*dx)
            aux(ix,iy,1) = tension_d1(ix) * surfactant_x
        end do
                
    end do  

    call film_x_velocity_apply_bcs(mx, my, mbc,  &
                                   aux(:,:,1))
    !<--- Film velocity field, x-component <---
    
    
    !---> Film velocity field, y-component --->
    ! Fill interior cells
    !$omp parallel do private(tension_d1, ix, surfactant_y)
    do iy=1,my
        tension_d1(1:mx) = q(1:mx,iy,2)
        call surface_tension_d1(tension_d1,mu,mx)
        
        do ix=1,mx
            surfactant_y = (q(ix,iy+1,2) - q(ix,iy-1,2)) / (2.d0*dy)
            aux(ix,iy,2) = tension_d1(ix) * surfactant_y
        end do
    end do
        
    call film_y_velocity_apply_bcs(mx, my, mbc,  &
                                   aux(:,:,2))
    !<--- Film velocity field, y-component <---


    ! Need film's Laplacian to compute surfactant velocity field
    call film_apply_bcs(mx, my, mbc,  &
                        q(:,:,1))
    call laplacian(mx, my, mbc, dx, dy,  &
                   q(:,:,1),  &
                   film_laplacian)
    call film_laplacian_apply_bcs(mx, my, mbc,  &
                                  film_laplacian)
    
    
    !---> Surfactant velocity field, y-component --->
    ! Fill interior cells
    !$omp parallel do private(ix, film_x, film_laplacian_x)
    do iy=1,my
        do ix=1,mx
            film_x           = (q(ix+1,iy,1)  - q(ix-1,iy,1))  / (2.d0*dx)
            film_laplacian_x = (film_laplacian(ix+1,iy) - film_laplacian(ix-1,iy)) / (2.d0*dx)
            aux(ix,iy,3) = q(ix,iy,1)**2/2.d0 * (-beta*film_x + kappa*film_laplacian_x)
        end do                    
    end do
    
    call surfactant_x_velocity_apply_bcs(mx, my, mbc,  &
                                         aux(:,:,3))
    !<--- Surfactant velocity field, x-component <---


    !---> Surfactant velocity field, y-component --->
    ! Fill interior cells
    !$omp parallel do private(ix, film_y, film_laplacian_y)
    do iy=1,my
        do ix=1,mx
            film_y = (q(ix,iy+1,1) -  q(ix,iy-1,1))  / (2.d0*dy)
            film_laplacian_y = (film_laplacian(ix,iy+1) - film_laplacian(ix,iy-1)) / (2.d0*dy)
            aux(ix,iy,4) = q(ix,iy,1)**2/2.d0 * (-beta*film_y + kappa*film_laplacian_y)
        end do
    end do
    
    call surfactant_y_velocity_apply_bcs(mx, my, mbc,  &
                                         aux(:,:,4))
    !<--- Surfactant velocity field, y-component <---

end subroutine b4step2
