subroutine film_integrate_slice( rhs, h, h_laplacian, beta, kappa, dx, dt, mx, mbc )

!---------------------------------------------------------------
! Solves a 1-dimensional problem for the approximately-factored
! thin film equation.
!
! The primary work of this routine is unpacking the sparsely
! stored Jacobian of the flux function into 'backward_euler',
! which is passed to LAPACK routine 'dgbsv' to solve the system.
!
! The first two rows of backward_euler are working space for
! dgbsv, and we do not fill them.  Rows 3 through 7 store the
! five diagonals of the system's "Jacobian" (which is actually
! an approximate factor of the full Jacobian for the 2D system).
! The column index of each element remains fixed in transition
! to the row-based form, as illustrated below for a 7x7 matrix:
!
!   Full matrix:
!
!         c1 b2 a3
!         d1 c2 b3 a4
!         e1 d2 c3 b4 a5
!            e2 d3 c4 b5 a6 
!               e3 d4 c5 b6 a7
!                  e4 d5 c6 b7
!                     e5 d6 c7
!
!
!   Rows 3 through 7 of backward_euler:
!
!          0  0 a3 a4 a5 a6 a7
!          0 b2 b3 b4 b5 b6 b7
!         c1 c2 c3 c5 c5 c6 c7
!         d1 d2 d3 d4 d5 d6  0
!         e1 e2 e3 e4 e5  0  0
!
!---------------------------------------------------------------


    implicit none


    !---> Variables --->

    !---- Inputs ----
    integer, intent(in) :: mx, mbc
    double precision, intent(in) :: h(1-mbc:mx+mbc), h_laplacian(1-mbc:mx+mbc)
    double precision, intent(in) :: beta, kappa, dx, dt
    !---- In/out ----
    double precision, intent(inout) :: rhs(mx)
    !---- Locals ----
    integer :: ipiv(mx), info, i, j, n_lower, n_upper
    parameter (n_lower=2, n_upper=2)
    double precision :: backward_euler(2*n_lower+1+n_upper, mx), sparse_flux(n_lower+n_upper,mx)

    !<--- Variables <---



    !---- Compute the flux operator ----
    call film_sparse_flux(sparse_flux, h, h_laplacian, beta, kappa, dx, dt, mx, mbc)



    !---- Define backward_euler ----
    !--------------------------------------------------------------------------------
    ! The body of the loop implements
    !
    !     backward_euler(3,j) =        dt/dx * (sparse_flux(1,j)                   )
    !     backward_euler(4,j) =        dt/dx * (sparse_flux(2,j) - sparse_flux(1,j))
    !     backward_euler(5,j) = 1.d0 + dt/dx * (sparse_flux(3,j) - sparse_flux(2,j))
    !     backward_euler(6,j) =        dt/dx * (sparse_flux(4,j) - sparse_flux(3,j))
    !     backward_euler(7,j) =        dt/dx * (                 - sparse_flux(4,j))
    !
    ! in a form that can be used with problems of other orders as well.
    !--------------------------------------------------------------------------------
    
    do j = 1, mx

        backward_euler(n_lower+1,j) = dt/dx * sparse_flux(1,j)
        
        
        do i = 2, n_lower+n_upper
            backward_euler(n_lower+i,j) = dt/dx * (sparse_flux(i,j) - sparse_flux(i-1,j))
        end do
        
        backward_euler(2*n_lower+1,j) = backward_euler(2*n_lower+1,j) + 1.d0
        
        
        backward_euler(2*n_lower+n_upper+1,j) = -dt/dx * sparse_flux(n_lower+n_upper,j)

    end do





    !---- Solve; rhs is overwritten with the result ----
    call dgbsv(mx, n_lower, n_upper, 1, backward_euler, 2*n_lower+n_upper+1, ipiv, rhs, mx, info)


end subroutine film_integrate_slice