subroutine film_parabolic_step(h, dh, h_newton, h_laplacian, tolerance, &
                               beta, kappa, dx, dy, dt, mx, my, mbc, success)

!-----------------------------------------------------
! Takes a time step with the parabolic portion of the 
! film equation, using Witelski's Newton ADI method.
!-----------------------------------------------------

    !$ use omp_lib
    implicit none


    !===> Variables ===>
    
    !---- Inputs ----
    integer, intent(in) :: mx, my, mbc
    double precision, intent(in) :: tolerance, beta, kappa, dx, dy, dt

    !---- In/out ----
    double precision, intent(inout) :: h(1-mbc:mx+mbc,1-mbc:my+mbc)

    !---- Output ----
    !-------------------------------------------------------------------
    ! The output values of arrays here aren't actually used elsewhere.
    ! The intent just specifies that their input values are irrelevant.
    !-------------------------------------------------------------------
    logical, intent(out) :: success
    double precision, intent(out), dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: dh, h_newton, h_laplacian

    !---- Globals ----
    integer :: newton_max_iter
    double precision :: newton_reduction_factor
    common /newton_block/ newton_max_iter, newton_reduction_factor

    !---- Locals ----
    double precision :: norm_dh, old_norm_dh
    integer :: iter, ix, iy, maxiter=10

    integer :: j_perm = 100
    !$ integer :: nthreads
    !$ common /omp_block/ nthreads

    !<==== Variables <===


    !$ call omp_set_num_threads(nthreads)

    print *, "Entering film_parabolic_step with dt = ", dt




    !===> Newton iteration ===>

    !---- Initialize ----
    !$omp parallel do private(ix)
    do iy = 1,my
        do ix = 1,mx
            h_newton(ix,iy) = h(ix,iy)
        end do
    end do
    
    success     = .false.
    norm_dh     = tolerance + 1.d0
    old_norm_dh = norm_dh
    iter = 0


    !---- Fill ghost cells of h; this is current until the routine terminates ----

    call extend_to_ghosts( h, 'even', 'periodic', mx, my, mbc )

    
    !---- Loop ----
    do while (norm_dh > tolerance)
        
        !---- Iteration count ----
        iter = iter + 1
        

        !---- Calculate the Laplacian of the current Newton iterate ----

        call extend_to_ghosts( h_newton, 'even', 'periodic', mx, my, mbc )
        call laplacian( h_laplacian, h_newton, dx, dy, mx, my, mbc )
        


        !---- Calculate the RHS for Newton's method ----

        call extend_to_ghosts( h_laplacian, 'even', 'periodic', mx, my, mbc )        
        call film_newton_rhs( dh, h_newton, h_laplacian, h, &
                              beta, kappa, dx, dy, dt, mx, my, mbc)
        
        
        !---- Horizontal solves ----
        !$omp parallel do
        do iy = 1,my
            call film_integrate_slice( dh(1:mx,iy), h_newton(1:mx,iy), h_laplacian(1:mx,iy), &
                                       beta, kappa, dx, dt, mx )
        end do
        
        
        !---- Vertical solves ----
        !$omp parallel do
        do ix = 1,mx
            call film_integrate_slice_periodic( dh(ix,1:my), h_newton(ix,1:my), h_laplacian(ix,1:my), &
                                                beta, kappa, dy, dt, my )
        end do
        
        
        !---- Update h_newton and norm_dh ----
        old_norm_dh = norm_dh
        norm_dh = 0.d0
        
        !$omp parallel do private(ix) reduction(max: norm_dh)
        do iy = 1,my
            do ix = 1,mx
              
                h_newton(ix,iy) = h_newton(ix,iy) + dh(ix,iy)
            
                norm_dh = max(norm_dh, abs(dh(ix,iy)))

            end do
        end do
                        

        !---- Give up if norm_dh gets too big ----
        if (norm_dh > 1e6) exit

        
        !---- If we've converged, then finish ----
        if (norm_dh < tolerance) then

            !$omp parallel do private(ix)
            do iy = 1,my
                do ix = 1,mx
                    h(ix,iy) = h_newton(ix,iy)
                end do
            end do

            success = .true.
            exit
        end if
        
        
        !---- If we've hit the max # of iterations, give up unless we're making sufficient progress ----
        if ( (iter >= newton_max_iter) .and. (norm_dh > newton_reduction_factor*old_norm_dh) ) exit
                                

    end do

    !<=== Newton iteration <===





!     contains
! 
!     subroutine permute (v)
!         implicit none
!         double precision :: v(1-mbc:mx+mbc,1-mbc:my+mbc)
!         double precision :: v_temp(my)
!         
!         do ix = 1, mx
!             v_temp(1:j_perm)  = v(ix, my-j_perm+1:my)
!             v_temp(j_perm+1:my) = v(ix, 1:my-j_perm)
!             v(ix,1:my) = v_temp(1:my)
!         end do
! 
!     end subroutine permute
! 
!     subroutine unpermute (v)
!         implicit none
!         double precision :: v(1-mbc:mx+mbc,1-mbc:my+mbc)
!         double precision :: v_temp(my)
! 
!         do ix = 1, mx
!             v_temp(my-j_perm+1:my) = v(ix, 1:j_perm)
!             v_temp(1:my-j_perm)    = v(ix, j_perm+1:my)
!             v(ix, 1:my) = v_temp(1:my)
!         end do
!         
!     end subroutine unpermute

end subroutine film_parabolic_step