subroutine film_sparse_flux( sparse_flux, h, h_laplacian, beta, kappa, dx, dt, mx, mbc )

!-------------------------------------------------------------
! This routine calculates the components in the 1-dimensional
! flux operator in an approximate factorization of the
! Jacobian for Newton's method.
!
! Factor of 1/2 comes from multiplicative factors.  Since this
! is  an interface-centered operator, that takes an average
! of cell-centered values from either side.
!-------------------------------------------------------------


    !---> Variables --->

    implicit none
    
    !---- Inputs ----
    integer, intent(in) :: mx, mbc
    double precision, intent(in) :: h(1-mbc:mx+mbc), h_laplacian(1-mbc:mx+mbc)
    double precision, intent(in) :: beta, kappa, dx, dt

    !---- Outputs ----
    double precision, intent(out) :: sparse_flux(4,mx)
    

    !---- Locals ----
    double precision :: h_face, h_x, h_laplacian_x, f, fd1
    integer :: j

    !<---= Variables <---=



    !---> Lower boundary --->
    
    
    !---- Zero the unused upper-left corner ----
    
    sparse_flux(1,1) = 0.d0
    
    
    
    !---- Interface 1 ----
    !--------------------------------------------------------------------
    ! We can impose the zero flux condition by setting these componennts
    ! to zero directly, but there are actually two things going on here.
    ! The multiplicative operator, seen below as 
    !              fd1*(-beta*h_x + kappa*h_laplacian_x),
    ! vanishes because h (the current Newton iterate) and h_laplacian
    ! are even across the boundary.  The other terms, coming from
    ! derivative stencils, will vanish because the function dh being 
    ! operated on is implicitly extended evenly across the boundary, so
    ! that dh(-1) = dh(2) and dh(0) = dh(1).
    !--------------------------------------------------------------------
    
    sparse_flux(2,1) = 0.d0
    sparse_flux(1,2) = 0.d0



    !---- Interface 2 ----
    !--------------------------------------------------------------------------
    ! The zero flux condition is visible here in the 2.d0*kappa/dx**3 in
    ! position (3,1), which would otherwise be 3.d0*kappa/dx**3.  The function
    ! that this flux operates on is implicitly extended evenly across the
    ! boundary, effectively setting dh(0) = dh(1).
    !--------------------------------------------------------------------------
        
    h_face        = (h(1) + h(2)) / 2.d0
    h_x           = (h(2) - h(1)) / dx
    h_laplacian_x = (h_laplacian(2) - h_laplacian(1)) / dx
    f             = h_face**3 / 3.d0
    fd1           = h_face**2

    sparse_flux(3,1) = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*( beta/dx + 2.d0*kappa/dx**3)
    sparse_flux(2,2) = fd1*(-beta*h_x + kappa*h_laplacian_x) * .5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
    sparse_flux(1,3) = f*(kappa/dx**3)

    !<--- Lower boundary <---



    !---> Interior interfaces --->

    do j = 3,mx-1
  
        h_face        = (h(j-1) + h(j)) / 2.d0
        h_x           = (h(j) - h(j-1)) / dx
        h_laplacian_x = (h_laplacian(j) - h_laplacian(j-1)) / dx
        f             = h_face**3 / 3.d0
        fd1           = h_face**2
            
        sparse_flux(4,j-2) = f*(-kappa/dx**3)
        sparse_flux(3,j-1) = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(beta/dx + 3.d0*kappa/dx**3)
        sparse_flux(2,j)   = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(-beta/dx - 3.d0*kappa/dx**3)
        sparse_flux(1,j+1) = f*(kappa/dx**3)

    end do

    !<--- Interior interfaces <---    



    !---> Upper boundary --->

    !---- Interface mx ----
    !------------------------------
    ! See comment for interface 2.
    !------------------------------
        
    h_face        = (h(mx-1) + h(mx)) / 2.d0
    h_x           = (h(mx) - h(mx-1)) / dx
    h_laplacian_x = (h_laplacian(mx) - h_laplacian(mx-1)) / dx
    f             = h_face**3 / 3.d0
    fd1           = h_face**2
    
    sparse_flux(4,mx-2) = f * (-kappa/dx**3)
    sparse_flux(3,mx-1) = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(beta/dx + 3.d0*kappa/dx**3)
    sparse_flux(2,mx)   = fd1*(-beta*h_x + kappa*h_laplacian_x)*.5d0 + f*(-beta/dx - 2.d0*kappa/dx**3)



    !---- Interface mx+1 ----
    !------------------------------
    ! See comment for interface 1.
    !------------------------------
    
    sparse_flux(4,mx-1) = 0.d0
    sparse_flux(3,mx)   = 0.d0

    
    
    !---- Zero the unused lower-right corner ----
    
    sparse_flux(4,mx) = 0.d0

    !<--- Upper boundary <---



end subroutine film_sparse_flux