subroutine surfactant_parabolic_step(g, h, g_newton, dg, tolerance, &
                                     delta, mu, dx, dy, dt, mx, my, mbc, success)

!----------------------------------------------------------
! Takes a time step with the parabolic portion of the 
! surfactant equation, using Witelski's Newton ADI method.
!----------------------------------------------------------

    !$ use omp_lib

    !===> Variables ===>

    implicit none

    !---- Inputs ----
    integer, intent(in) :: mx, my, mbc
    double precision, intent(in) :: h(1-mbc:mx+mbc,1-mbc:my+mbc), tolerance, delta, mu, dx, dy, dt

    !---- In/out ----
    double precision, intent(inout) :: g(1-mbc:mx+mbc,1-mbc:my+mbc)

    !---- Output ----
    !-------------------------------------------------------------
    ! The output values of the arrays here are not actually used.
    ! The intent just indicates that their input values do not
    ! matter.
    !-------------------------------------------------------------
    double precision, intent(out), dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: g_newton, dg
    logical, intent(out) :: success
    
    !---- Globals ----
    integer :: newton_max_iter
    double precision :: newton_reduction_factor
    common /newton_block/ newton_max_iter, newton_reduction_factor

    !---- Locals ----
    double precision :: norm_dg, old_norm_dg
    integer :: iter, ix, iy, maxiter=10, imax, jmax
    double precision :: g_temp(1-mbc:mx+mbc,1-mbc:my+mbc), & 
                        dg_temp(1-mbc:mx+mbc,1-mbc:my+mbc), temp_norm
    !$ integer :: nthreads
    !$ common /omp_block/ nthreads
    
    !<==== Variables <===



    !---- Initialize ----
    !$ call omp_set_num_threads(nthreads)
    print *, "Entering surfactant_parabolic_step with dt = ", dt
    
    !$omp parallel do private(ix)
    do iy = 1,my
        do ix = 1,mx
            g_newton(ix,iy) = g(ix,iy)
        end do
    end do
    
    success  = .false.
    norm_dg     = tolerance + 1.d0
    old_norm_dg = norm_dg
    iter = 0


    !---- Fill ghost cells; these are current throughout the routine ----

    call extend_to_ghosts( g, 'even', 'periodic', mx, my, mbc )
    call extend_to_ghosts( h, 'even', 'periodic', mx, my, mbc )
    


    !===> Newton iterations ===>
    do while (norm_dg > tolerance)
        
        iter = iter + 1
        
        !---- Calculate RHS for Newton's method; store in dg ----
        
        call extend_to_ghosts( g_newton, 'even', 'periodic', mx, my, mbc )
        
        call surfactant_newton_rhs( dg, g_newton, g, h, &
                                    delta, mu, dx, dy, dt, mx, my, mbc )

                                    
        
        
        !---- Solve horizontal problems, overwriting dg ----
        !$omp parallel do
        do iy = 1,my
            call surfactant_integrate_slice( dg(1:mx,iy), g_newton(1:mx,iy), h(1:mx,iy), &
                                             delta, mu, dx, dt, mx )

        end do

        
        !---- Solve vertical problems, overwriting dg ----
        !$omp parallel do
        do ix = 1,mx
            call surfactant_integrate_slice_periodic( dg(ix,1:my), g_newton(ix,1:my), h(ix,1:my), &
                                                      delta, mu, dy, dt, my )
        end do
        
        
        !---- Update g_newton, norm_dg, and old_norm_dg ----
        old_norm_dg = norm_dg
        norm_dg = 0.d0

        !$omp parallel do private(ix) reduction(max: norm_dg)
        do iy = 1,my
            do ix = 1,mx

                g_newton(ix,iy) = g_newton(ix,iy) + dg(ix,iy)

                norm_dg = max(norm_dg, abs(dg(ix,iy)))

            end do
        end do
        

        !---- Give up if norm_dg is too big ----
        if (norm_dg > 1e6) exit


        !---- If we've converged, then finish ----
        if (norm_dg < tolerance) then

            !$omp parallel do private(ix)
            do iy = 1,my
                do ix = 1,mx
                    g(ix,iy) = g_newton(ix,iy)
                end do
            end do

            success = .true.
            exit
        end if
        
        !---- If we've hit the max # of iterations, give up unless we're making sufficient progress ----
        if ( (iter >= newton_max_iter) .and. (norm_dg > newton_reduction_factor*old_norm_dg) ) exit

    end do
    !<=== Newton iterations <===




    contains
    
    subroutine check_the_fucking_difference ( a, b )
        implicit none
        double precision :: fucking_difference
        double precision, dimension(1-mbc:mx+mbc,1-mbc:my+mbc) :: a, b
        
        do iy = 1, my
            do ix = 1, mx
                fucking_difference = max( fucking_difference, abs(a(ix,iy) - b(ix,iy)))
            end do
        end do
        
        if (fucking_difference < 1.d-14) then
            print *, "There is no fucking difference."
        else
            print *, "The goddamn fucking difference is ", fucking_difference
        end if
        
    end subroutine check_the_fucking_difference


end subroutine surfactant_parabolic_step