subroutine forward_euler_1d(q, diffusivity, dx, dt, mx)

!-----------------------------------------------------
! Performs a tridiagonal solve for 1D backward Euler,
!     (I - dt*D)qnew = qold,
! where D = d/dx(k d/dx).
!-----------------------------------------------------


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: diffusivity, dx, dt
    
    !---- Input/output ----
    double precision, intent(inout) :: q(0:mx+1)
    

    !---- Locals ----
    integer :: i
    double precision :: flux_low, flux_high

    !<=== Variables <===



    flux_low = -diffusivity * (q(1) - q(0)) / dx

    do i=1,mx
        flux_high = -diffusivity * (q(i+1) - q(i)) / dx
        q(i)      = q(i) - dt/dx * (flux_high - flux_low)
        flux_low  = flux_high
    end do



end subroutine forward_euler_1d





subroutine backward_euler_1d(q, diffusivity, dx, dt, mx)

!-----------------------------------------------------
! Performs a tridiagonal solve for 1D backward Euler,
!     (I - dt*D)qnew = qold,
! where D = d/dx(k d/dx).
!-----------------------------------------------------


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: diffusivity, dx, dt
    
    !---- Input/output ----
    double precision, intent(inout) :: q(0:mx+1)
    

    !---- Locals ----
    integer :: info
    double precision :: c
    double precision, dimension(mx-1) :: diagm1, diagp1
    double precision, dimension(mx) :: diag

    !<=== Variables <===


    !---- Set up the tridiagonal matrix, excluding boundaries ----
    c = dt/dx**2

    diagm1(:) = -c*diffusivity
    diagp1(:) = -c*diffusivity
    diag(:)   = 1.d0 + c * 2.d0*diffusivity



    !---- Include boundary effects ----
    q(1)  = q(1)  + c*diffusivity*q(0)
    q(mx) = q(mx) + c*diffusivity*q(mx+1)


 
    !---- Update q ----
    call dgtsv(mx, 1, diagm1, diag, diagp1, q(1:mx), mx, info)    


end subroutine backward_euler_1d





subroutine crank_nicolson_1d(q, future_bcs, diffusivity, dx, dt, mx)


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: future_bcs(2), diffusivity, dx, dt

    !---- Input/output ----
    double precision, intent(inout) :: q(0:mx+1)

    !<=== Variables <===
    
    
    call forward_euler_1d( q, diffusivity, dx, dt/2, mx )
    
    q(0)    = future_bcs(1)
    q(mx+1) = future_bcs(2)
    call backward_euler_1d( q, diffusivity, dx, dt/2, mx )

end subroutine crank_nicolson_1d





subroutine tga_1d(q, future_bcs, diffusivity, dx, dt, mx)


    implicit none


    !===> Variables ===>

    !---- Inputs ----
    integer, intent(in) :: mx
    double precision, intent(in) :: future_bcs(2), diffusivity, dx, dt

    !---- Input/output ----
    double precision, intent(inout) :: q(0:mx+1)


    !---- Locals ----
    double precision :: a, r1, r2, r3, old_bcs(2)

    !<=== Variables <===
    
    
    !---- Calculate lengths of substeps ----
    a = 2.d0 - sqrt(2.d0) - 1d-8
    r1 = 1.d0-a
    r2 = (2.d0*a-1.d0) / (a + sqrt(a**2 - 4.d0*a + 2.d0))
    r3 = (2.d0*a-1.d0) / (a - sqrt(a**2 - 4.d0*a + 2.d0))
    
    
    !---- Copy initial BCs ----
    old_bcs(1) = q(0)
    old_bcs(2) = q(mx+1)
    
    
    !---- Steps ----
    call forward_euler_1d( q, diffusivity, dx, r1*dt, mx )
    
    q(0)    = old_bcs(1) + (r1 + r2)*(future_bcs(1) - old_bcs(1))
    q(mx+1) = old_bcs(2) + (r1 + r2)*(future_bcs(2) - old_bcs(2))
    call backward_euler_1d( q, diffusivity, dx, r2*dt, mx )

    q(0)    = future_bcs(1)
    q(mx+1) = future_bcs(2)
    call backward_euler_1d( q, diffusivity, dx, r3*dt, mx )


end subroutine tga_1d
