subroutine crank_nicolson_3d_split ( q, diffusivity, t, dt, cell_width, x_low, n_cells, n_ghosts )

    !$ use omp_lib

    implicit none

    !==== Variables ====

    !---- Inputs ----
    integer, intent(in) :: n_cells(3), n_ghosts
    double precision, intent(in) :: diffusivity, t, dt, cell_width(3), x_low(3)

    !---- Input/output ----
    double precision, intent(inout) :: q( 1-n_ghosts : n_cells(1)+n_ghosts,  &
                                          1-n_ghosts : n_cells(2)+n_ghosts,  &
                                          1-n_ghosts : n_cells(3)+n_ghosts )
                                          
    !---- Locals ----
    integer :: ix, iy, iz, i_dim
    double precision :: x, y, z, coordinate, dx, dy, dz
    double precision :: new_bcs(2), t_new
    integer :: nx, ny, nz
    double precision :: q_true, q_true_here
    double precision :: q_xx, q_xx_terms, q_yy, q_yy_terms, q_zz, q_zz_terms
    
    
    
    !---- Statement functions ----
    coordinate(i_dim,ix) = x_low(i_dim) + (ix-0.5d0)*cell_width(i_dim)



    !---- Extract grid dimensions ----
    nx = n_cells(1)
    ny = n_cells(2)
    nz = n_cells(3)
    dx = cell_width(1)
    dy = cell_width(2)
    dz = cell_width(3)


    !---- Time at end of step ----
    t_new = t + dt



    !|\"""""""""""""""""|\
    !| >    x-sweeps    | >
    !|/_________________|/

    do iz = 1, nz
        z = coordinate(3,iz)
        
        !$omp parallel do private(y, x, q_true_here, q_xx_terms, q_yy, q_zz, q_xx, new_bcs)
        do iy = 1, ny
            y = coordinate(2,iy)
            

            !===> Lower boundary data ===>

            x = coordinate(1,0)

            !---- Stabilize the lower initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_xx_terms  = explicit_dxx_terms( q(1:3,iy,iz), dx )
            q_yy        = ( q_true(x,y-dy,z,t_new) - 2*q_true_here + q_true(x,y+dy,z,t_new) ) / dy**2
            q_zz        = ( q_true(x,y,z-dz,t_new) - 2*q_true_here + q_true(x,y,z+dz,t_new) ) / dz**2
            
            q(0,iy,iz) = ( q_true_here - dt * diffusivity * ( q_xx_terms + q_yy + q_zz ) )  &
                            / (1.d0 + 2 * dt * diffusivity / dx**2 )
                            
            !---- Estimate the lower final BC ----
            q_xx       = one_sided_dxx( q(0:3, iy, iz), dx )
            new_bcs(1) = q(0,iy,iz) + dt * diffusivity * q_xx

            !<=== Lower boundary data <===
            
            
            
            !===> Upper boundary data ===>
            
            x = coordinate(1,nx+1)
            
            !---- Stabilize the upper initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_xx_terms  = explicit_dxx_terms( q(nx:nx-2:-1,iy,iz), dx )
            q_yy        = ( q_true(x,y-dy,z,t_new) - 2*q_true_here + q_true(x,y+dy,z,t_new)) / dy**2
            q_zz        = ( q_true(x,y,z-dz,t_new) - 2*q_true_here + q_true(x,y,z+dz,t_new)) / dz**2

            q(nx+1,iy,iz) = ( q_true_here - dt * diffusivity * ( q_xx_terms + q_yy + q_zz ) )  &
                              / (1.d0 + 2 * dt * diffusivity / dx**2 )
                              
            !---- Estimate the upper final BC ----
            q_xx       = one_sided_dxx( q(nx+1:nx-2:-1, iy, iz), dx )
            new_bcs(2) = q(nx+1,iy,iz) + dt * diffusivity * q_xx
            
            !<=== Upper boundary data <===



            !---- Solve 1d problem ----
            call crank_nicolson_1d( q(0:nx+1,iy,iz), new_bcs, diffusivity, dx, dt, nx )
        
        end do
    end do
    
    ! /|"""""""""""""""""/|
    !< |    x-sweeps    < |
    ! \|_________________\|





    !|\"""""""""""""""""|\
    !| >    y-sweeps    | >
    !|/_________________|/
    
    do iz = 1, nz
        z = coordinate(3,iz)
        
        !$omp parallel do private(x, y, q_true_here, q_yy_terms, q_zz, q_yy, new_bcs)
        do ix = 1, nx
            x = coordinate(1,ix)
            
            !===> Lower boundary data ===>

            y = coordinate(2,0)

            !---- Stabilize the lower initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_yy_terms  = explicit_dxx_terms( q(ix,1:3,iz), dy )
            q_zz        = ( q_true(x,y,z-dz,t_new) - 2*q_true_here + q_true(x,y,z+dz,t_new)) / dz**2
            
            q(ix,0,iz) = ( q_true_here - dt * diffusivity * ( q_yy_terms + q_zz ) )  &
                            / (1.d0 + 2 * dt * diffusivity / dy**2 )
                            
            !---- Estimate the lower final BC ----
            q_yy       = one_sided_dxx( q(ix, 0:3, iz), dy )
            new_bcs(1) = q(ix,0,iz) + dt * diffusivity * q_yy

            !<=== Lower boundary data <===
            
            
            
            !===> Upper boundary data ===>
            
            y = coordinate(2,ny+1)
            
            !---- Stabilize the upper initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_yy_terms  = explicit_dxx_terms( q(ix,ny:ny-2:-1,iz), dy )
            q_zz        = ( q_true(x,y,z-dz,t_new) - 2*q_true_here + q_true(x,y,z+dz,t_new) ) / dz**2

            q(ix,ny+1,iz) = ( q_true_here - dt * diffusivity * ( q_yy_terms + q_zz ) )  &
                              / (1.d0 + 2 * dt * diffusivity / dy**2 )
                              
            !---- Estimate the upper final BC ----
            q_yy       = one_sided_dxx( q(ix, ny+1:ny-2:-1, iz), dy )
            new_bcs(2) = q(ix,ny+1,iz) + dt * diffusivity * q_yy
            
            !<=== Upper boundary data <===
            
            
        
            !---- Solve 1d problem ----
            call crank_nicolson_1d( q(ix,0:ny+1,iz), new_bcs, diffusivity, dy, dt, ny )
        
        end do
    end do
    
    ! /|"""""""""""""""""/|
    !< |    y-sweeps    < |
    ! \|_________________\|
    
    
    

    !|\"""""""""""""""""|\
    !| >    z-sweeps    | >
    !|/_________________|/

    do iy = 1, ny
        y = coordinate(2,iy)
        
        !$omp parallel do private(x, z, q_true_here, q_zz_terms, new_bcs)
        do ix = 1, nx
            x = coordinate(1,ix)
        
            !===> Lower boundary data ===>

            z = coordinate(3,0)

            !---- Stabilize the lower initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_zz_terms  = explicit_dxx_terms( q(ix,iy,1:3), dz )            
        
            q(ix,iy,0) = ( q_true_here - dt * diffusivity * q_zz_terms )  &
                            / (1.d0 + 2 * dt * diffusivity / dz**2 )
                        
            !---- Lower final BC is exact ----
            new_bcs(1) = q_true_here

            !<=== Lower boundary data <===
        
        
        
            !===> Upper boundary data ===>
        
            z = coordinate(3,nz+1)
        
            !---- Stabilize the upper initial BC ----
            q_true_here = q_true(x,y,z,t_new)
            q_zz_terms  = explicit_dxx_terms( q(ix,iy,nz:nx-2:-1), dz )

            q(ix,iy,nz+1) = ( q_true_here - dt * diffusivity * q_zz_terms )  &
                              / (1.d0 + 2 * dt * diffusivity / dz**2 )
                          
            !---- Final upper BC is exact ----
            new_bcs(2) = q_true_here
        
            !<=== Upper boundary data <===



            !---- Solve 1d problem ----
            call crank_nicolson_1d( q(ix,iy,0:nz+1), new_bcs, diffusivity, dz, dt, nz )

        end do
    end do
    
    ! /|"""""""""""""""""/|
    !< |    z-sweeps    < |
    ! \|_________________\|
    


    contains
    
    function one_sided_dxx( q_slice, dx )
        double precision, intent(in) :: dx, q_slice(4)
        double precision :: one_sided_dxx
        
        one_sided_dxx = ( 2*q_slice(1) - 5*q_slice(2) + 4*q_slice(3) - q_slice(4) ) / dx**2
    end function one_sided_dxx
    
    function explicit_dxx_terms( q_slice, dx )
        double precision, intent(in) :: dx, q_slice(3)
        double precision :: explicit_dxx_terms
        
        explicit_dxx_terms = ( -5*q_slice(1) + 4*q_slice(2) - q_slice(3) ) / dx**2
        
    end function explicit_dxx_terms
    

end subroutine crank_nicolson_3d_split