program dimension_splitting_3d
    
    implicit none
    
    !==== Variables ====
    integer :: n_cells(3), n_ghosts=1
    double precision :: t, dt, t_limits(2)
    double precision :: x_low(3), x_high(3), dx(3)

    double precision :: diffusivity
    common /split_block/ diffusivity

    double precision, allocatable :: q(:,:,:)

    integer :: ix, iy, iz, i_dim
    double precision :: x, y, z, coordinate
    double precision :: q_true
    double precision :: error_value
    
    
    
    !---- Statement functions for coordinate values ----
    coordinate(i_dim,ix) = x_low(i_dim) + (ix-.5d0)*dx(i_dim)
    
    
    !===> Read simulation parameters from file ===>

    open(7, file="SimulationData.txt", status="old")

    read(7,*) x_low
    read(7,*) x_high
    read(7,*) t_limits

    read(7,*) dt
    read(7,*) n_cells
    
    read(7,*) diffusivity
    
    do i_dim = 1,3
        dx(i_dim) = (x_high(i_dim) - x_low(i_dim)) / n_cells(i_dim)
    end do

    close(7)
    
    !<=== Read simulation parameters from file <===
    
    
    
    !===> Allocate and initialize solution ===>
    
    allocate( q( 1-n_ghosts : n_cells(1)+n_ghosts,  &
                 1-n_ghosts : n_cells(2)+n_ghosts,  &
                 1-n_ghosts : n_cells(3)+n_ghosts )   )

    do iz = 1, n_cells(3)
        z = coordinate(3,iz)
        
        do iy = 1, n_cells(2)
            y = coordinate(2,iy)
            
            do ix = 1, n_cells(1)
                x = coordinate(1,ix)
                
                q(ix,iy,iz) = q_true( x, y, z, t_limits(1) )
            end do
        end do
    end do

    !<=== Allocate and initialize solution <===
    

    
    
    !===> Time-stepping ===>
    
    t = t_limits(1)
    
    do while ( abs(t - t_limits(2)) > 1.d-8 )
        
        
        !---- Adjust time step if necessary ----
        if (t+dt > t_limits(2)) then
            dt = t_limits(2) - t
        end if
        
        
        !---- Advance the solution ----
        call crank_nicolson_3d_split( q, diffusivity, t, dt, dx, x_low, n_cells, n_ghosts )
        
        
        !---- Update time ----
        t = t + dt

    end do
    
    !<=== Time-stepping <===

    
    
    print *, q_true(x_low(1), 0.d0, 0.d0, t_limits(2))
    
    !---- Check numerical error ----
    
    error_value = 0.d0
    
    do iz = 1, n_cells(3)
        z = coordinate(3,iz)
        
        do iy = 1, n_cells(2)
            y = coordinate(2,iy)
            
            do ix = 1, n_cells(1)
                x = coordinate(1,ix)
                
                error_value = max(error_value, abs(q(ix,iy,iz) - q_true(x,y,z,t)))
            end do  
        end do
    end do
    
        
    
    open( 7, file="NumericalError.txt" )
    write(7,*) error_value
    close(7)
    
    
    
    
end program dimension_splitting_3d


subroutine scan(q, n_cells, n_ghosts)
    implicit none
    integer, intent(in) :: n_cells(3), n_ghosts
    double precision, intent(in) :: q( 1-n_ghosts : n_cells(1)+n_ghosts,  &
                                          1-n_ghosts : n_cells(2)+n_ghosts,  &
                                          1-n_ghosts : n_cells(3)+n_ghosts )
    integer :: ix, iy, iz
    
    do iz = 1, n_cells(3)
        do iy = 1, n_cells(2)
            do ix = 1, n_cells(1)
                if ( abs(q(ix,iy,iz)) > 1.d2 ) then
                    print *, "Warning: q exceeds tolerance at ", ix, iy, iz
                end if
            end do
        end do
    end do
    
end subroutine scan



function q_true ( x, y, z, t )
    implicit none
    double precision, intent(in) :: x, y, z, t
    double precision :: q_true
    double precision :: diffusivity, pi=acos(-1.d0), t0=.5d0, tau
    common /split_block/ diffusivity
    
    tau = t + t0
    
    q_true = 1.d0/(4*pi*diffusivity*tau)**(1.5d0) * exp( -(x**2+y**2+z**2) / (4*diffusivity*tau) )
    
!     q_true = exp( -3*pi**2 * diffusivity * t) * sin(pi*x) * sin(pi*y) * sin(pi*z)
    
end function q_true