from numpy import *
from IntegrationOperators import *
    



class ForwardAdvectionBCEstimator:
    
    velocity = 0.
    dx = 0.
    dt = 0.
    n  = 10

    
    def __init__(self, velocity, dx, dt, n, order=2 ):
        
        self.velocity = velocity
        self.dx = dx
        self.dt = dt
        self.n  = n
        
        self.courant = velocity*dt/dx


        E = zeros( (2,self.n) )
        
        if order==1:
            if self.velocity>0.:
                E[0,0] = -self.courant
            elif self.velocity<0.:
                E[1,-1] = self.courant

        elif order==2:
            if self.velocity>0.:
                E[0,0:2] = -self.courant * array([2., -.5])
            elif self.velocity<0.:
                E[1,-2:] = -self.courant * array([.5, -2.])

        self.interior = E


        B = zeros( (2,2) )
        
        if order==1:
            if velocity>0.:
                B[0,0] = 1. + self.courant
            elif velocity<0.:
                B[1,1] = 1. - self.courant
        
        elif order==2:
            if velocity>0.:
                B[0,0] = 1. - self.courant * (-1.5)
            elif velocity<0.:
                B[1,1] = 1. - self.courant * 1.5
        
        self.oldBoundary = B
    
        
        self.newBoundary = zeros( (2,2) )
    
    
    def __call__( self, q, bc_old ):

        return dot( self.interior, q ) + dot( self.oldBoundary, bc_old )
    




class BackwardAdvectionBCEstimator:
    
    velocity = 0.
    dx = 0.
    dt = 0.
    n  = 10

    
    def __init__(self, velocity, dx, dt, n):
        
        self.velocity = velocity
        self.dx = dx
        self.dt = dt
        self.n  = n
                
        self.courant = velocity*dt/dx

        
        E = zeros( (2,self.n) )
        
        # if self.velocity > 0.:
        #     E[0,:2] = array([2., -.5]) * self.courant / (1 + self.courant*1.5)
        # elif self.velocity < 0.:
        #     E[1,-2:] = array([.5, -2.]) * self.courant / (1. - 1.5*self.courant)    

        if self.velocity > 0.:
            E[0,:2] = array([1., 0.]) * self.courant / (1 + self.courant)
        elif self.velocity < 0.:
            E[1,-2:] = array([0., -1.]) * self.courant / (1. - self.courant)    
            
        self.interior = E
    
            
        self.oldBoundary = zeros( (2,2) )
    
            
        B = zeros( (2,2) )
        
        if self.velocity > 0.:
            B[0,0] = 1. / (1. + self.courant)
        elif self.velocity < 0.:
            B[1,1] = 1. / (1. - self.courant)
        
        self.newBoundary = B
        
        
    def __call__( self, q, bc_new ):
        
        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )
    




class StabilizedAdvectionBCEstimator:
    
    velocity = 0.
    dx = 0.
    dt = 0.
    dt_to_end = 0.
    n  = 10

    
    def __init__(self, velocity, dx, dt, dt_to_end, n):
        
        self.velocity  = velocity
        self.dx        = dx
        self.dt        = dt
        self.dt_to_end = dt_to_end
        self.n         = n
                
        self.courant = velocity*dt/dx

        
        BackEst = BackwardAdvectionBCEstimator( self.velocity, self.dx, self.dt_to_end, self.n )
        FwdEst  = ForwardAdvectionBCEstimator( self.velocity, self.dx, self.dt, self.n )
                    
        self.interior = FwdEst.interior + dot( FwdEst.oldBoundary, BackEst.interior )
        
        self.oldBoundary = zeros( (2,2) )
    
        self.newBoundary = dot( FwdEst.oldBoundary, BackEst.newBoundary )
        
    
    def __call__( self, q, bc_new ):

        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )





class ForwardDiffusionBCEstimator:

    diffusivity = 0.
    dx = 0.
    dt = 0.
    n  = 10


    def __init__(self, diffusivity, dx, dt, n):

        self.diffusivity = diffusivity
        self.dx = dx
        self.dt = dt
        self.n  = n

        self.stab_rat = diffusivity*dt/dx**2


        E = zeros( (2,self.n) )

        E[0,:3]  = self.stab_rat * array([-5., 4., -1.])
        E[1,-3:] = self.stab_rat * array([-1., 4., -5.])

        # # First order
        # E[0,:3]  = self.stab_rat * array([-2., 1., 0.])
        # E[1,-3:] = self.stab_rat * array([0., 1., -2.])


        self.interior = E


        B = zeros( (2,2) )

        B[0,0] = 1. + self.stab_rat*2.
        B[1,1] = 1. + self.stab_rat*2.

        
        # # First order
        # B[0,0] = 1. + self.stab_rat
        # B[1,1] = 1. + self.stab_rat

        self.oldBoundary = B


        self.newBoundary = zeros( (2,2) )

    def __call__( self, q, bc_old ):
        
        return dot( self.interior, q ) + dot( self.oldBoundary, bc_old )





class BackwardDiffusionBCEstimator:

    diffusivity = 0.
    dx = 0.
    dt = 0.
    n  = 10


    def __init__(self, diffusivity, dx, dt, n):

        self.diffusivity = diffusivity
        self.dx = dx
        self.dt = dt
        self.n  = n

        self.stab_rat = diffusivity*dt/dx**2


        E = zeros( (2,self.n) )

        E[0,:3]  = -self.stab_rat * array([-5., 4., -1.]) / (1. + self.stab_rat*2.)
        E[1,-3:] = -self.stab_rat * array([-1., 4., -5.]) / (1. + self.stab_rat*2.)

        self.interior = E


        self.oldBoundary = zeros( (2,2) )


        B = zeros( (2,2) )
        B[0,0] = 1. / (1. + self.stab_rat*2.)
        B[1,1] = 1. / (1. + self.stab_rat*2.)

        self.newBoundary = B

        
    def __call__( self, q, bc_new ):

        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )





class StabilizedDiffusionBCEstimator:

    diffusivity = 0.
    dx = 0.
    dt = 0.
    dt_to_end = 0.
    n  = 10


    def __init__(self, diffusivity, dx, dt, dt_to_end, n):

        self.diffusivity = diffusivity
        self.dx        = dx
        self.dt        = dt
        self.dt_to_end = dt_to_end
        self.n         = n

        self.stab_rat = diffusivity*dt/dx**2


        BackEst = BackwardDiffusionBCEstimator( self.diffusivity, self.dx, self.dt_to_end, self.n )
        FwdEst = ForwardDiffusionBCEstimator( self.diffusivity, self.dx, self.dt, self.n )

        self.interior = FwdEst.interior + dot( FwdEst.oldBoundary, BackEst.interior )


        self.oldBoundary = zeros( (2,2) )


        self.newBoundary = dot( FwdEst.oldBoundary, BackEst.newBoundary )


    def __call__( self, q, bc_new ):

        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )






class BackwardAdvDiffBCEstimator:

    def __init__(self, velocity, diffusivity, dx, dt, n):

        self.velocity    = velocity
        self.diffusivity = diffusivity

        self.dx = dx
        self.dt = dt
        self.n  = n

        self.courant = velocity*dt/dx
        self.diff_ratio = diffusivity*dt/dx**2


        E = zeros( (2,self.n) )

        E[0,0:3] = ( self.courant * array([1., 0., 0.]) - self.diff_ratio * array([-5., 4., -1.]) )  \
                     / (1. + self.courant + 2.*self.diff_ratio)
        
        E[1,-3:] = ( self.courant * array([0., 0., -1.]) - self.diff_ratio * array([-1., 4., -5.]) )  \
                     / (1. - self.courant + 2.*self.diff_ratio)
            
        self.interior = E


        self.oldBoundary = zeros( (2,2) )


        B = zeros( (2,2) )

        B[0,0] = 1. / (1. + self.courant + 2.*self.diff_ratio)
        B[1,1] = 1. / (1. - self.courant + 2.*self.diff_ratio)

        self.newBoundary = B


    def __call__( self, q, bc_new ):

        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )




class BackwardAdvDiffUpstreamBCEstimator:

    def __init__(self, velocity, diffusivity, dx, dt, n):

        self.velocity    = velocity
        self.diffusivity = diffusivity

        self.dx = dx
        self.dt = dt
        self.n  = n

        self.courant = velocity*dt/dx
        self.diff_ratio = diffusivity*dt/dx**2


        E = zeros( (2,self.n) )

        if velocity > 0.:
            E[0,0:3] = ( self.courant * array([1., 0., 0.]) - self.diff_ratio * array([-5., 4., -1.]) )  \
                         / (1. + self.courant + 2.*self.diff_ratio)
        else:
            E[1,-3:] = ( self.courant * array([0., 0., -1.]) - self.diff_ratio * array([-1., 4., -5.]) )  \
                         / (1. - self.courant + 2.*self.diff_ratio)

        self.interior = E


        self.oldBoundary = zeros( (2,2) )


        B = zeros( (2,2) )

        if velocity > 0.:
            B[0,0] = 1. / (1. + self.courant + 2.*self.diff_ratio)
        else:
            B[1,1] = 1. / (1. - self.courant + 2.*self.diff_ratio)

        self.newBoundary = B


    def __call__( self, q, bc_new ):

        return dot( self.interior, q ) + dot( self.newBoundary, bc_new )





if __name__ == "__main__":
    
    from pylab import *
    import time
    # 
    # 
    # class SineMode:
    #     
    #     diffusivity = 0.
    #     mode = 1
    # 
    #     def __init__(self, diffusivity, mode):
    #         self.diffusivity = diffusivity
    #         self.mode = mode
    # 
    #     def __call__(self, x, t):
    #         return exp(-(self.mode*pi)**2 * self.diffusivity * t) * sin(self.mode*pi*x)
    # 
    # 
    # 
    # x = linspace(-.8, .7, 101)
    # x_interior = x[1:-1]
    # x_boundary = x[[0,-1]]
    # 
    # dx = x[1] - x[0]
    # n = len(x_interior)
    # t_final = .4
    # dt = .04
    # diffusivity = .1
    # 
    # 
    # q_true = SineMode(diffusivity, 2)
    # q = q_true(x_interior,0.)
    # time.sleep(.2)
    # 
    # clf()
    # p, = plot(x_interior, q, 'ro')
    # p_true, = plot( x_interior, q_true(x_interior, 0.))
    # 
    # 
    # 
    # CN = CrankNicolsonOperator( diffusivity, dx, dt/3., n )
    # 
    # FwdEst  = ForwardDiffusionBCEstimator( diffusivity, dx, dt/3., n )
    # BackEst = BackwardDiffusionBCEstimator( diffusivity, dx, dt/2, n )
    # StabEst = StabilizedDiffusionBCEstimator( diffusivity, dx, dt/2, dt, n )
    # 
    # BackEst13 = BackwardDiffusionBCEstimator( diffusivity, dx, dt/3., n )
    # BackEst23 = BackwardDiffusionBCEstimator( diffusivity, dx, 2.*dt/3., n )
    # 
    # 
    # null_bc = -100*ones(2)
    # 
    # t = 0.
    # 
    # while (t < t_final-1e-12):
    # 
    #     bc_old = q_true( x_boundary, t )
    #     bc_new = q_true( x_boundary, t+dt )
    #     
    #     bc_mid2 = FwdEst( q, bc_old )
    #     
    #     q = CN( q, bc_old, bc_mid2 )
    #     
    #     bc_mid1 = BackEst23( q, bc_new )
    #     bc_mid2 = FwdEst( q, bc_mid1)
    #     
    #     q = CN( q, bc_mid1, bc_mid2 )
    # 
    #     bc_mid1 = BackEst13( q, bc_new )
    #     
    #     q = CN( q, bc_mid1, bc_new)
    # 
    # 
    #     
    # 
    #     t += dt
    # 
    #     p.set_ydata(q)
    #     p_true.set_ydata( q_true(x_interior,t) )
    #     draw()
    #     time.sleep(.1)








    # class AdvectingGaussian:
    # 
    #     def __init__(self, velocity):
    #         self.velocity = velocity
    # 
    #     def __call__(self, x, t):
    #         return exp(-200*(x-velocity*t)**2)
    # 
    # 
    # x_low  = 0.0
    # x_high = 0.7
    # 
    # x = linspace(x_low, x_high, 101)
    # x_interior = x[1:-1]
    # x_boundary = x[[0,-1]]
    # 
    # dx = x[1] - x[0]
    # n = len(x_interior)
    # t_final = 1.3
    # dt = .025
    # velocity = 0.5
    # 
    # 
    # q_true = AdvectingGaussian(velocity)
    # q = q_true(x_interior,0.)
    # 
    # clf()
    # p, = plot(x_interior, q, 'ro')
    # p_true, = plot( x_interior, q_true(x_interior, 0.))
    # time.sleep(.1)
    # 
    # null_bc = zeros(2)
    # 
    # LW = LaxWendroffOperator( velocity, dx, dt/2, n )
    # 
    # FwdEst  = ForwardAdvectionBCEstimator( velocity, dx, dt/2, n )
    # BackEstHalf = BackwardAdvectionBCEstimator( velocity, dx, dt/2, n )
    # BackEstFull = BackwardAdvectionBCEstimator( velocity, dx, dt, n )
    # StabEst = StabilizedAdvectionBCEstimator( velocity, dx, dt/2, dt, n )
    # 
    # 
    # t = 0.
    # 
    # while (t < t_final-1e-12):
    #     
    #     bc_old = q_true( x_boundary, t )
    #     bc_new = q_true( x_boundary, t+dt )
    # 
    #     q = LW(q, bc_old)
    # 
    #     bc_mid = BackEstHalf( q, bc_new )
    #         
    #     q = LW(q, bc_mid)
    #     
    #     t += dt
    #     
    #     p.set_ydata(q)
    #     p_true.set_ydata( q_true(x_interior,t) )
    #     draw()
    #     raw_input("blah")
    #     # time.sleep(.2)





    class AdvectingGaussian:
    
        def __init__( self, velocity, diffusivity, t0 ):
            self.velocity = velocity
            self.diffusivity = diffusivity
            self.t0 = t0
        
        def __call__( self, x, t ):
            tau = t + self.t0
            return 1./sqrt(4.*pi*self.diffusivity*tau) * exp( -(x-velocity*tau)**2 / (4.*self.diffusivity*tau) )
    
    
    
    x_low  = 0.
    x_high = 1.0
    
    x = linspace(x_low, x_high, 101)
    x_interior = x[1:-1]
    x_boundary = x[[0,-1]]
    
    dx = x[1] - x[0]
    n = len(x_interior)
    t_final = 1.3
    dt = .018
    velocity = 0.5
    diffusivity = .05
    
    
    q_true = AdvectingGaussian(velocity, diffusivity, .3)
    q = q_true(x_interior,0.)
    
    clf()
    p, = plot(x_interior, q, 'ro')
    p_true, = plot( x_interior, q_true(x_interior, 0.))
    time.sleep(.1)
    
    null_bc = zeros(2)
    
    LW = LaxWendroffOperator( velocity, dx, dt, n )
    CN = CrankNicolsonOperator( diffusivity, dx, dt, n )
    
    BackEstAdvDiff   = BackwardAdvDiffBCEstimator( velocity, diffusivity, dx, dt, n )
    BackEstAdvection = BackwardAdvectionBCEstimator( velocity, dx, dt, n )
    BackEstDiffusion = BackwardDiffusionBCEstimator( diffusivity, dx, dt, n )
    FwdEstDiffusion  = ForwardDiffusionBCEstimator( diffusivity, dx, dt, n )
    
    
    
    t = 0.
    
    while (t < t_final-1e-12):
        
        bc_old = q_true( x_boundary, t )
        bc_new = q_true( x_boundary, t+dt )

        q = LW( q, bc_old )
        bc1 = BackEstDiffusion( q, bc_new )
        q = CN( q, bc1, bc_new )
    
    
        # bc1 = BackEstAdvDiff( q, bc_new )
        # bc2 = FwdEstDiffusion( q, bc1 )
        # 
        # q = CN( q, bc1, bc2 )
        #     
        # bc1 = BackEstAdvection( q, bc_new )
        # 
        # q = LW( q, bc1 )
        
        t += dt
        
        p.set_ydata(q)
        p_true.set_ydata( q_true(x_interior,t) )
        draw()
        # raw_input("Press Enter to advance")
        time.sleep(.05)